// vim:fdm=syntax
// by tuberry

import St from 'gi://St';
import GObject from 'gi://GObject';
import Clutter from 'gi://Clutter';

import * as Main from 'resource:///org/gnome/shell/ui/main.js';
import * as PanelMenu from 'resource:///org/gnome/shell/ui/panelMenu.js';
import * as PopupMenu from 'resource:///org/gnome/shell/ui/popupMenu.js';

import { Field } from './const.js';
import { Lyric } from './lyric.js';
import { xnor, noop, homolog } from './util.js';
import { MprisPlayer as Mpris } from './mpris.js';
import { DesktopPaper, PanelPaper } from './paper.js';
import { SwitchItem, MenuItem, TrayIcon } from './menu.js';
import { Fulu, ExtensionBase, Destroyable, manageSource, omit, connect, getSelf, _ } from './fubar.js';

class LyricButton extends PanelMenu.Button {
    static {
        GObject.registerClass(this);
    }

    constructor(callback) {
        super(0.5);
        this._onXButtonClick = callback;
        let box = new St.BoxLayout({ style_class: 'panel-status-menu-box' });
        box.add_child(new TrayIcon('lyric-symbolic', true));
        this.set_paper = x => x && box.add_child(x);
        this.add_child(box);
    }

    vfunc_event(event) {
        if(event.type() === Clutter.EventType.BUTTON_PRESS && event.get_button() >>> 1 === 4) {
            this._onXButtonClick();
            return Clutter.EVENT_STOP;
        }
        return super.vfunc_event(event);
    }
}

class DesktopLyric extends Destroyable {
    constructor(gset) {
        super();
        this._buildWidgets();
        this._bindSettings(gset);
    }

    _buildWidgets() {
        this._lyric = new Lyric();
        this._mpris = new Mpris();
        this._src = manageSource(this, () => omit(this, '_mpris', '_lyric', '_paper'), {
            sync: [clearTimeout, x => setTimeout(x, 500)],
            tray: [() => { this.systray = false; }, () => { this.systray = true; }],
            play: [clearInterval, x => x && setInterval(() => this.setPosition(this._paper._moment + this._span + 0.625), this._span)],
        });
    }

    _bindSettings(gset) {
        this._fulu = new Fulu({}, gset, this);
        this._fulu.attach({
            mini:  [Field.MINI, 'boolean'],
            drag:  [Field.DRAG, 'boolean'],
            index: [Field.TIDX, 'uint'],
            path:  [Field.PATH, 'string'],
            span:  [Field.SPAN, 'uint'],
        }, this);
        connect(this, [this._mpris, 'update', this._update.bind(this),
            'closed', (_p, closed) => { this.closed = closed; },
            'status', (_p, status) => { this.playing = status === 'Playing'; },
            'seeked', (_p, position) => this.setPosition(position / 1000)]);
    }

    set path(path) {
        this._lyric.location = path;
    }

    set mini(mini) {
        this._mini = mini;
        if(this._paper) omit(this, 'playing', '_paper');
        if(mini) {
            this._paper = new PanelPaper(this._fulu);
            this._btn?.set_paper(this._paper);
            this._menus?.drag.hide();
        } else {
            this._paper = new DesktopPaper(this._fulu);
            this._menus?.drag.show();
        }
        this.loadLyric();
    }

    set systray(systray) {
        if(xnor(systray, this._btn)) return;
        if(systray) {
            this._btn = Main.panel.addToStatusArea(getSelf().uuid, new LyricButton(() => this.syncPosition()),
                this._index ? 0 : 5, ['left', 'center', 'right'][this._index ?? 0]);
            this._addMenuItems();
            this._btn.visible = this._showing;
            if(this._mini) this.mini = this._mini;
        } else {
            if(this._mini) omit(this, '_paper');
            omit(this, '_btn', '_menus');
        }
    }

    set index(index) {
        if(this._index === index) return;
        this._index = index;
        this._src.tray.refreshSource();
    }

    set drag(drag) {
        this._drag = drag;
        this._menus?.drag.setToggleState(drag);
    }

    set span(span) {
        this._span = span;
        if(this._src.play._delegate) this.playing = true;
    }

    set playing(playing) {
        this._updateViz();
        this._src.play.refreshSource(playing && this._paper);
    }

    set closed(closed) {
        this._showing = !closed;
        if(closed) this.clearLyric();
        if(this._btn) this._btn.visible = !closed;
    }

    async syncPosition() {
        if(this._syncing) return;
        this._syncing = true;
        let pos = await this._mpris.getPosition() / 1000;
        for(let i = 0; pos && (pos === this._pos || !this._length || this._length - pos < 2000) && i < 7; i++) { // HACK: workaround for stale positions from buggy NCM mpris when changing songs
            await new Promise(resolve => this._src.sync.refreshSource(resolve));
            pos = await this._mpris.getPosition() / 1000;
        }
        this.setPosition((this._pos = pos) + 50);
        this._syncing = false;
    }

    _update(_player, song, length) {
        if(homolog(this._song, song, undefined, (x, y, k) => k === 'lyric' ? x?.length === y?.length : x === y)) {
            this.syncPosition();
        } else {
            this._title = Lyric.format(song, ' - ', '/');
            this._length = length;
            this._song = song;
            this.loadLyric();
        }
    }

    setPosition(pos) {
        this._paper.moment = pos;
    }

    loadLyric(reload) {
        if(!this._song) return;
        this.setLyric('');
        if(this._song.lyric !== null) this.setLyric(this._song.lyric);
        else this._lyric.find(this._song, reload).then(x => this.setLyric(x)).catch(noop);
    }

    setLyric(text) {
        if(!this._paper) return;
        this._paper.span = this._length ?? 0;
        this._paper.text = text;
        this._paper.song = this._mini ? this._title : '';
        this.playing = this._mpris.status === 'Playing';
        this.syncPosition();
    }

    clearLyric() {
        this.playing = false;
        this._paper.clear();
        this._song = null;
    }

    _updateViz() {
        let viz = this._mpris.status === 'Playing' && !this._menus?.hide.state;
        if(this._paper && this._paper.visible ^ viz) this._paper.visible = viz;
    }

    _addMenuItems() {
        this._menus = {
            hide:   new SwitchItem(_('Invisiblize'), false, () => this._updateViz()),
            mini:   new SwitchItem(_('Minimize'), this._mini, x => this._fulu.set('mini', x, this)),
            drag:   new SwitchItem(_('Mobilize'), this._drag, x => this._fulu.set('drag', x, this)),
            sep0:   new PopupMenu.PopupSeparatorMenuItem(),
            reload: new MenuItem(_('Redownload'), () => this.loadLyric(true)),
            resync: new MenuItem(_('Resynchronize'), () => this.syncPosition()),
            sep1:   new PopupMenu.PopupSeparatorMenuItem(),
            prefs:  new MenuItem(_('Settings'), () => getSelf().openPreferences()),
        };
        Object.values(this._menus).forEach(x => this._btn.menu.addMenuItem(x));
    }
}

export default class Extension extends ExtensionBase { $klass = DesktopLyric; }
